(*******************************************************************************

TLinePrinter Version 2.0
8/8/96 - 7/20/97
Copyright (c) 1996-1997 Bill Menees

This is a non-visual VCL component that encapsulates the QRPrinter object.

It requires long strings which means it can't be used with Delphi 1.0.
It works fine with Delphi 2 & 3 and C++Builder 1.0.

For more information see the included Doc file.

*******************************************************************************)

{ $LONGSTRINGS ON }
unit LinePrnt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,
  Forms, Dialogs, Printers, StdCtrls, ExtCtrls, QuickRep;

const
     {In Pixels}
     DefaultDPI = 300;
     DefaultBorderWidth = 2;
     {In Inches}
     DefaultPhysicalPageHeightIn = 11.0;
     DefaultPhysicalPageWidthIn = 8.5;
     DefaultAvailablePageHeightIn = 10.5;
     DefaultAvailablePageWidthIn = 8.0;
     DefaultGutterLeftIn = 0.25;
     DefaultGutterTopIn = 0.25;
     {In Millimeters}
     DefaultPhysicalPageHeightMm = 297.0;
     DefaultPhysicalPageWidthMm = 210.0;
     DefaultAvailablePageHeightMm = 284.0;
     DefaultAvailablePageWidthMm = 198.0;
     DefaultGutterLeftMm = 6.0;
     DefaultGutterTopMm = 6.0;

     {These are expanded only in
     Headers, Footers, and Tables.}
     LineField = '{$LINE}';
     PageField = '{$PAGE}';
     DateField = '{$DATE}';
     TimeField = '{$TIME}';
     TitleField = '{$TITLE}';

type
  ELinePrinter = class(EPrinter);

  {These are declared so you can tell at a glance what
  a property or function's return value is used for.}
  TMeasurement = Single;
  TPixels = Cardinal;

  TPrntProgDlg = class(TForm)
    Bevel: TBevel;
    lblStatus: TLabel;
    lblTitle: TLabel;
    lblPageNumDesc: TLabel;
    lblPageNumber: TLabel;
  private
    { Private declarations }
    procedure WMNCHitTest(var Msg: TWMNCHitTest); message WM_NCHITTEST;
  public
    { Public declarations }
  end;

  TMeasureUnit = (muInches, muMillimeters);
  TLineSpacing = (lsHalfSpace, lsSingleSpace, lsSingleAndAHalf, lsDoubleSpace);
  TPageBorderStyle = (pbTop, pbBottom, pbLeft, pbRight);
  TPageBorders = set of TPageBorderStyle;

  TLinePrinter = class(TComponent)
  private
    { Private declarations }
    fPrinter: TQRPrinter;
    fCanvas: TCanvas;
    fPrntProgDlg: TPrntProgDlg;

    fLineNumber: Cardinal;
    fPageNumber: Cardinal;
    fMarginTop, fMarginBottom, fMarginLeft, fMarginRight: TMeasurement;
    fMeasureUnit: TMeasureUnit;
    fLineSpacing: TLineSpacing;
    fTabSize: Cardinal;
    fWordWrap: Boolean;
    fLines: TStrings;
    fAlignment: TAlignment;
    fHeader, fFooter: String;
    fPageBorders: TPageBorders;
    fTextMetrics: TTextMetric;
    fShowProgress: Boolean;
    fTableFormat: String;
    fDefaultColWidth: TMeasurement;
    fBorderOffset: TMeasurement;
    fHeaderFormat, fFooterFormat: String;
    fTableGrid: Boolean;
    fPrintToFile: Boolean;
    fFileName: String;
    fPrintingToFile: Boolean;
    fFileVar: TextFile;
    fFriendlyFooter: String;
    fFriendlyHeader: String;
    fTokenSeparator: Char;

    {These X,Y are relative to the printable space.
    They should normally be bounded by the Margins.
    So 0,0 is the left,top corner of the printable space.
    fCurrentY is negative only when printing the header.}
    fCurrentX, fCurrentY: Integer;
    fLineSpace: TPixels;

    fOnBeginDoc: TNotifyEvent;
    fOnEndDoc: TNotifyEvent;
    fOnAbortDoc: TNotifyEvent;
    fOnNewLine: TNotifyEvent;
    fOnNewPage: TNotifyEvent;

    function GetAborted: Boolean;
    function GetFont: TFont;
    function GetOrientation: TPrinterOrientation;
    function GetAvailablePageHeight: TMeasurement;
    function GetAvailablePageWidth: TMeasurement;
    function GetPageNumber: Cardinal;
    function GetPrinting: Boolean;
    function GetTitle: String;
    function GetGutterTop: TMeasurement;
    function GetGutterBottom: TMeasurement;
    function GetGutterLeft: TMeasurement;
    function GetGutterRight: TMeasurement;
    function GetLines: TStrings;
    
    procedure SetMarginTop(Value: TMeasurement);
    procedure SetMarginBottom(Value: TMeasurement);
    procedure SetMarginLeft(Value: TMeasurement);
    procedure SetMarginRight(Value: TMeasurement);
    procedure SetMeasureUnit(Value: TMeasureUnit);
    procedure SetLineSpacing(Value: TLineSpacing);
    procedure SetTabSize(Value: Cardinal);
    procedure SetWordWrap(Value: Boolean);
    procedure SetLines(Value: TStrings);
    procedure SetAlignment(Value: TAlignment);
    procedure SetHeader(Value: String);
    procedure SetFooter(Value: String);
    procedure SetPageBorders(Value: TPageBorders);

    procedure SetFont(Value: TFont);
    procedure SetOrientation(Value: TPrinterOrientation);
    procedure SetTitle(Value: String);
    procedure SetShowProgress(Value: Boolean);
    procedure SetTableFormat(Value: String);
    procedure SetDefaultColWidth(Value: TMeasurement);
    procedure SetBorderOffset(Value: TMeasurement);
    procedure SetHeaderFormat(Value: String);
    procedure SetFooterFormat(Value: String);

    procedure SetPrintToFile(Value: Boolean);
    procedure SetFileName(Value: String);

    function GetPhysicalPageHeight: TMeasurement;
    function GetPhysicalPageWidth: TMeasurement;
    function GetPrintableHeight: TMeasurement;
    function GetPrintableWidth: TMeasurement;
    function PixelPrintWidth: TPixels;
    function PixelPrintHeight: TPixels;
    function StartingLeft: TPixels;
    function StartingRight: TPixels;
    function StartingTop: TPixels;
    function StartingBottom: TPixels;

    procedure SetFriendlyFooter(Value: String);
    procedure SetFriendlyHeader(Value: String);
    function GetFriendlyFooter: String;
    function GetFriendlyHeader: String;

  protected
    { Protected declarations }
    procedure SetPixelsPerInch;
    procedure SplitLineAndPrint(const Line: String);
    procedure DoNewPageProcessing;
    procedure UpdateProgressDlg(const Status: String);
    function GetClippedLine(const Line: String; const Width: TPixels): String;
    function MeasureUnitsToPixelsH(const M: TMeasurement): TPixels;
    function MeasureUnitsToPixelsV(const M: TMeasurement): TPixels;
    function PixelsToMeasureUnitsH(const P: TPixels): TMeasurement;
    function PixelsToMeasureUnitsV(const P: TPixels): TMeasurement;
    function ExpandLogicalFields(S: String): String;

  public
    { Public declarations }
    property Aborted: Boolean read GetAborted;
    property Canvas: TCanvas read fCanvas write fCanvas;
    property LineNumber: Cardinal read fLineNumber;
    //This is the Printer.PageHeight/Width property converted to TMeasurement.
    //It's the largest available printable space per page.
    property AvailablePageHeight: TMeasurement read GetAvailablePageHeight;
    property AvailablePageWidth: TMeasurement read GetAvailablePageWidth;
    //This is how large the piece of paper physically is.
    property PhysicalPageHeight: TMeasurement read GetPhysicalPageHeight;
    property PhysicalPageWidth: TMeasurement read GetPhysicalPageWidth;
    //This is the printable area determined by the margins.
    property PrintableHeight: TMeasurement read GetPrintableHeight;
    property PrintableWidth: TMeasurement read GetPrintableWidth;
    property PageNumber: Cardinal read GetPageNumber;
    property Printing: Boolean read GetPrinting;

    property GutterTop: TMeasurement read GetGutterTop;
    property GutterBottom: TMeasurement read GetGutterBottom;
    property GutterLeft: TMeasurement read GetGutterLeft;
    property GutterRight: TMeasurement read GetGutterRight;

    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;

    procedure AbortDoc;
    procedure BeginDoc;
    procedure EndDoc;
    function NewPage: Cardinal;
    function NewLine: Cardinal;
    function PrevLine: Boolean;

    procedure WriteLine(const Line: String);
    procedure WriteLineRight(const Line: String);
    procedure WriteLineCenter(const Line: String);
    procedure WriteTableLine(const Line: String);
    procedure PrintLines(const LinesAsTable: Boolean);
    procedure Refresh;
    procedure ExpandFriendlyFormat(UserFmt: String; AsHeader: Boolean);

  published
    { Published declarations }
    property MarginTop: TMeasurement read fMarginTop write SetMarginTop;
    property MarginBottom: TMeasurement read fMarginBottom write SetMarginBottom;
    property MarginLeft: TMeasurement read fMarginLeft write SetMarginLeft;
    property MarginRight: TMeasurement read fMarginRight write SetMarginRight;
    property MeasureUnit: TMeasureUnit read fMeasureUnit write SetMeasureUnit default muInches;
    property LineSpacing: TLineSpacing read fLineSpacing write SetLineSpacing default lsSingleSpace;
    property TabSize: Cardinal read fTabSize write SetTabSize default 8;
    property WordWrap: Boolean read fWordWrap write SetWordWrap default True;
    property Lines: TStrings read GetLines write SetLines;
    property Alignment: TAlignment read fAlignment write SetAlignment default taLeftJustify;
    property Header: String read fHeader write SetHeader nodefault;
    property HeaderFormat: String read fHeaderFormat write SetHeaderFormat;
    property Footer: String read fFooter write SetFooter nodefault;
    property FooterFormat: String read fFooterFormat write SetFooterFormat;
    property PageBorders: TPageBorders read fPageBorders write SetPageBorders default [];
    property ShowProgress: Boolean read fShowProgress write SetShowProgress default False;
    property Font: TFont read GetFont write SetFont;
    property Orientation: TPrinterOrientation read GetOrientation write SetOrientation default poPortrait;
    property Title: String read GetTitle write SetTitle nodefault;
    property TableFormat: String read fTableFormat write SetTableFormat;
    property DefaultColWidth: TMeasurement read fDefaultColWidth write SetDefaultColWidth;
    property BorderOffset: TMeasurement read fBorderOffset write SetBorderOffset;
    property TableGrid: Boolean read fTableGrid write fTableGrid default False;
    property PrintToFile: Boolean read fPrintToFile write SetPrintToFile default False;
    property FileName: String read fFileName write SetFileName;
    property TokenSeparator: Char read fTokenSeparator write fTokenSeparator default '|';

    property OnBeginDoc: TNotifyEvent read fOnBeginDoc write fOnBeginDoc;
    property OnEndDoc: TNotifyEvent read fOnEndDoc write fOnEndDoc;
    property OnAbortDoc: TNotifyEvent read fOnAbortDoc write fOnAbortDoc;
    property OnNewLine: TNotifyEvent read fOnNewLine write fOnNewLine;
    property OnNewPage: TNotifyEvent read fOnNewPage write fOnNewPage;
  end;

procedure Register;
function ReplaceSubString(OldSubStr, NewSubStr, S: String): String;
procedure ParseFormatToken(var CurToken: String; var CurAlignment: TAlignment;
          var CurWidth: TMeasurement; const DefaultColWidth: TMeasurement);
procedure TokenizeString(const S: String; TokenSeparator: Char;
          Tokens: TStringList);
function StripBackToWhiteSpace(const S: String): String;
function ExpandTabsAsSpaces(const S: String; const TabSize: Integer): String;
function GenSpace(const Size: Integer): String;

{$R PrntProg.dfm}

implementation

procedure Register;
begin
     RegisterComponents('Menees', [TLinePrinter]);
end;

{=============================================================================}
{ Non-methods that may prove useful elsewhere.                                }
{=============================================================================}

function ReplaceSubString(OldSubStr, NewSubStr, S: String): String;
var
   P: Cardinal;
begin
     {Currently, this routine is terribly inefficient since Pos
     always starts back at the beginning of the string.
     Eventually, I hope to replace this with a Knuth-Morris-Pratt
     based search and replace that starts at a specified location.
     This works for now though for what I need.}

     {First, make sure old isn't contained in new.
     This gets around the infinite loop situation.
     If old is in new, we just return S unmodified.}
     P:=Pos(OldSubStr, NewSubStr);
     if P = 0 then
     begin
          P:=Pos(OldSubStr, S);
          while P > 0 do
          begin
               S:=Copy(S, 1, P-1)+NewSubStr+Copy(S, P+Length(OldSubStr), Length(S));
               P:=Pos(OldSubStr, S);
          end;
     end;
     Result:=S;
end;

procedure ParseFormatToken(var CurToken: String; var CurAlignment: TAlignment; var CurWidth: TMeasurement; const DefaultColWidth: TMeasurement);
begin
     if CurToken = '' then CurToken:='<'+FloatToStr(DefaultColWidth);
     if Length(CurToken) = 1 then
        if (CurToken[1] in ['<', '^', '>']) then CurToken:=CurToken+FloatToStr(DefaultColWidth);

     {Alignment}
     case CurToken[1] of
          '<': begin
                    CurAlignment:=taLeftJustify;
                    Delete(CurToken, 1, 1);
               end;
          '^': begin
                    CurAlignment:=taCenter;
                    Delete(CurToken, 1, 1);
               end;
          '>': begin
                    CurAlignment:=taRightJustify;
                    Delete(CurToken, 1, 1);
               end;
     else
         CurAlignment:=taLeftJustify;
     end;

     {Width}
     try
        CurWidth:=StrToFloat(CurToken);
     except
           on EConvertError do CurWidth:=DefaultColWidth;
     end;
end;

procedure TokenizeString(const S: String; TokenSeparator: Char; Tokens: TStringList);
var
   i, Len: Cardinal;
   CurToken: String;
begin
     Tokens.Clear;
     CurToken:='';
     Len:=Length(S);
     for i:=1 to Len do
     begin
          if S[i] = TokenSeparator then
          begin
               Tokens.Add(CurToken);
               CurToken:='';
          end
          else
              CurToken:=CurToken+S[i];
     end;
     Tokens.Add(CurToken);
end;

function StripBackToWhiteSpace(const S: String): String;
var
   i, Len, Mark: Cardinal;
begin
     Mark:=0;
     Len:=Length(S);
     for i:=Len downto 1 do
     begin
          if S[i] in [#0..#32] then
          begin
               Mark:=i;
               Break;
          end;
     end;

     if Mark > 0 then Result:=Copy(S, 1, Mark)
     {If there is nowhere to break, just return the whole line.}
     else Result:=S;
end;

function ExpandTabsAsSpaces(const S: String; const TabSize: Integer): String;
var
   i, Len, Size: Cardinal;
   Buffer: String;
begin
     {TabStr:='';
     for i:=1 to TabSize do TabStr:=TabStr+' ';}

     Buffer:='';
     Len:=Length(S);
     for i:=1 to Len do
     begin
          if S[i]=#9 then
          begin
               Size:=TabSize-(Length(Buffer) mod TabSize);
               Buffer:=Buffer+GenSpace(Size);
          end
          else Buffer:=Buffer+S[i];
     end;
     Result:=Buffer;
end;

function GenSpace(const Size: Integer): String;
var
   Str: String;
begin
     Str:='';
     while Length(Str) < Size do Str:=Str+' ';
     GenSpace:=Str;
end;

{=============================================================================}
{ Private stuff for TPrntProgDlg.                                             }
{=============================================================================}

procedure TPrntProgDlg.WMNCHitTest(var Msg: TWMNCHitTest);
begin
     {Don't let them resize the progress dialog.}
     inherited;
     with Msg Do
          if (Result = HTTop) or
             (Result = HTTopLeft) or
             (Result = HTTopRight) or
             (Result = HTLeft) or
             (Result = HTRight) or
             (Result = HTBottom) or
             (Result = HTBottomLeft) or
             (Result = HTBottomRight) then Result:=HTNowhere;
end;

{=============================================================================}
{ Public stuff for TLinePrinter.                                              }
{=============================================================================}

constructor TLinePrinter.Create(Owner: TComponent);
begin
     inherited Create(Owner);
     {Make sure things don't blow up if there is no printer.}
     fPrinter:=nil;
     try
        fPrinter:=QRPrinter;
        fCanvas:=fPrinter.Canvas;
     except
           on EPrinter do
           begin
                if csDesigning in ComponentState then
                   MessageDlg('You must have at least one printer installed to create and use all of the features of this component.',
                                   mtWarning, [mbOk], 0);
           end;
     end;
     {Note: This is created as a TStringList
     but declared as a TStrings. This is to
     maintain a consistent look with other
     VCL components.  TStrings is used as a
     visible outer layer while TStringList
     is used internally for storage.}
     fLines := nil;
     fLines := TStringList.Create;
     if fLines = nil then raise Exception.Create('Unable to create Lines property');
     
     {Make this explicitly nil so UpdateProgressDlg
     can tell if it needs to Create or Free itself.}
     fPrntProgDlg := nil;

     fCurrentX:=0;
     fCurrentY:=0;
     fLineNumber:=0;
     fPageNumber:=0;
     fPrintingToFile:=False;
     //It's hard to zero a non-typed variable...
     FillChar(fFileVar, sizeof(fFileVar), #0);

     Font.Name := 'Courier New';
     Font.Size := 10;
     Font.Style:= [];

     LineSpacing:=lsSingleSpace;
     TabSize:=8;
     WordWrap:=True;
     Alignment:=taLeftJustify;
     PageBorders:=[];
     Orientation:=poPortrait;
     ShowProgress:=False;
     Header:='';
     HeaderFormat:='';
     Footer:='';
     FooterFormat:='';
     TableFormat:='';
     Title:='';
     MarginTop:=GutterTop;
     MarginBottom:=GutterBottom;
     MarginLeft:=GutterLeft;
     MarginRight:=GutterRight;
     BorderOffset:=0;
     DefaultColWidth:=0;
     MeasureUnit:=muInches;
     TableGrid:=False;
     PrintToFile:=False;
     TokenSeparator:='|';
end;

destructor TLinePrinter.Destroy;
begin
     fLines.Free;
     inherited Destroy;
end;

procedure TLinePrinter.AbortDoc;
begin
     try
        UpdateProgressDlg('Aborting Printing');
        try
           if PrintToFile then
           begin
                CloseFile(fFileVar);
                FillChar(fFileVar, sizeof(fFileVar), #0);
                fPrintingToFile:=False;
           end
           else
           begin
                fPrinter.Cancel;
           end;
        except
              on EInOutError do raise ELinePrinter.Create('Unable to abort printing to file.');
              on EPrinter do raise ELinePrinter.Create('Unable to abort printing.');
        end;
     finally
            UpdateProgressDlg('');
     end;
     {Fire the event handler if it exists.}
     if Assigned(fOnAbortDoc) then fOnAbortDoc(Self);
end;

procedure TLinePrinter.BeginDoc;
begin
     {Fire the event handler if it exists.}
     if Assigned(fOnBeginDoc) then fOnBeginDoc(Self);
     try
        SetPixelsPerInch;
        if PrintToFile then
        begin
             AssignFile(fFileVar, FileName);
             Rewrite(fFileVar);
             fPrintingToFile:=True;
        end
        else
        begin
             fPrinter.BeginDoc;
             //Make sure the font gets sized correctly for the page.
             GetTextMetrics(Canvas.Handle, fTextMetrics);
             SetPixelsPerInch;
        end;
     except
           on E: EInOutError do raise ELinePrinter.Create(E.Message);
           on EPrinter do raise ELinePrinter.Create('Unable to begin printing.');
     end;
     UpdateProgressDlg('Preparing to Print');

     fPageNumber:=1;
     {Make sure the new page processing fires on BeginDoc.}
     DoNewPageProcessing;
end;

procedure TLinePrinter.EndDoc;
begin
     try
        UpdateProgressDlg('Finished Printing');
        try
           if PrintToFile then
           begin
                CloseFile(fFileVar);
                FillChar(fFileVar, sizeof(fFileVar), #0);
                fPrintingToFile:=False;
           end
           else
           begin
                fPrinter.EndDoc;
           end;
        except
              on EInOutError do raise ELinePrinter.Create('Unable to finish printing to file.');
              on EPrinter do raise ELinePrinter.Create('Unable to finish printing.');
        end;
     finally
            UpdateProgressDlg('');
     end;
     {Fire the event handler if it exists.}
     if Assigned(fOnEndDoc) then fOnEndDoc(Self);
end;

function TLinePrinter.NewPage: Cardinal;
begin
     try
        if fPrintingToFile then
           Writeln(fFileVar, #12)
        else
            fPrinter.NewPage;
        Inc(fPageNumber);
     except
           on EInOutError do raise ELinePrinter.Create('Unable to print a new page.');
           on EPrinter do raise ELinePrinter.Create('Unable to print a new page.');
     end;
     DoNewPageProcessing;
     Result:=PageNumber;
end;

function TLinePrinter.NewLine: Cardinal;
begin
     fCurrentX:=0;
     fCurrentY:=fCurrentY+fLineSpace;

     {See if the entire next line will fit.}
     if not fPrintingToFile and ((fCurrentY+fLineSpace) >= PixelPrintHeight) then
         NewPage
     else
     begin
          if fPrintingToFile then Writeln(fFileVar);
          Inc(fLineNumber);
     end;
     {Fire the event handler if it exists.}
     if Assigned(fOnNewLine) then fOnNewLine(Self);
     Result:=LineNumber;
end;

//This function returns whether it was successful.
function TLinePrinter.PrevLine: Boolean;
begin
     Result:=False;
     if (fCurrentY >= fLineSpace) and not fPrintingToFile then
     begin
          fCurrentX:=0;
          fCurrentY:=fCurrentY-fLineSpace;
          Dec(fLineNumber);
          Result:=True;
     end;
end;

procedure TLinePrinter.WriteLine(const Line: String);
var
   LineWidth: TPixels;
   Buffer: String;
begin
     if Pos(#9, Line)>0 then Buffer:=ExpandTabsAsSpaces(Line, TabSize)
     else Buffer:=Line;

     try
        LineWidth:=Canvas.TextWidth(Buffer);
     except
           on EPrinter do LineWidth:=0;
     end;
     if LineWidth > PixelPrintWidth then
     begin
          if WordWrap then SplitLineAndPrint(Buffer)
          else WriteLine(GetClippedLine(Buffer, PixelPrintWidth));
     end
     else
     begin
          case Alignment of
               taRightJustify: fCurrentX := PixelPrintWidth-LineWidth;
               taCenter: fCurrentX := (PixelPrintWidth-LineWidth) shr 1;
          else
              fCurrentX:=0;
          end;
          {Make sure we don't write off the end of the page.}
          if (fCurrentY+fLineSpace) >= PixelPrintHeight then
             if not fPrintingToFile then NewPage;
          {Now print the line.}
          try
             if fPrintingToFile then //Put spaces in for alignment purposes.
                Writeln(fFileVar, GenSpace(fCurrentX div Canvas.TextWidth(' '))+Buffer)
             else
                 Canvas.TextOut(StartingLeft+fCurrentX, StartingTop+fCurrentY, Buffer);
          except
                on EPrinter do ;
                on EInOutError do ;
          end;
          NewLine;
     end;
end;

procedure TLinePrinter.WriteLineRight(const Line: String);
var
   OldAlign: TAlignment;
begin
     OldAlign:=Alignment;
     try
        Alignment:=taRightJustify;
        WriteLine(Line);
     finally
            Alignment:=OldAlign;
     end;
end;

procedure TLinePrinter.WriteLineCenter(const Line: String);
var
   OldAlign: TAlignment;
begin
     OldAlign:=Alignment;
     try
        Alignment:=taCenter;
        WriteLine(Line);
     finally
            Alignment:=OldAlign;
     end;
end;

procedure TLinePrinter.WriteTableLine(const Line: String);
var
   FormatTokens, LineTokens: TStringList;
   i, CurWidth, LeftPos, X, Y, Offset: Integer;
   FloatCurWidth: TMeasurement;
   CurAlignment: TAlignment;
   CurToken: String;
begin
     FormatTokens:=TStringList.Create;
     LineTokens:=TStringList.Create;

     try
        TokenizeString(TableFormat, '|', FormatTokens);
        TokenizeString(Line, '|', LineTokens);

        fCurrentX:=StartingLeft;
        for i:=0 to FormatTokens.Count-1 do
        begin
             {Get the Width and Alignment from the current column format.}
             CurToken:=FormatTokens[i];
             ParseFormatToken(CurToken, CurAlignment, FloatCurWidth, DefaultColWidth);
             CurWidth:=MeasureUnitsToPixelsH(FloatCurWidth);

             {Now get a line token even if it's blank.}
             if i < LineTokens.Count then CurToken:=LineTokens[i]
             else CurToken:='';

             //Expand logical field names (e.g. {$LINE}).
             {The '{$' check is just to speed things up.}
             if Pos('{$', CurToken) > 0 then
                CurToken:=ExpandLogicalFields(CurToken);

             {Get just what will fit in the current column.}
             CurToken:=GetClippedLine(CurToken, CurWidth);

             try
                {Figure out where the X position will be in the current column.}
                case CurAlignment of
                     taCenter: LeftPos:=(CurWidth-Canvas.TextWidth(CurToken)) shr 1;
                     taRightJustify: LeftPos:=CurWidth-Canvas.TextWidth(CurToken);
                else
                    LeftPos:=0;
                end;
                {We try to offset the text so it's not right on a border.}
                Offset:=3*DefaultBorderWidth*(GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSX) div DefaultDPI);
                if Canvas.TextWidth(CurToken) < (CurWidth-Offset) then
                   if CurAlignment = taRightJustify then LeftPos:=LeftPos - Offset
                   else if CurAlignment = taLeftJustify then LeftPos:=LeftPos + Offset;

                {Print out the current token.}
                if fPrintingToFile then
                   Write(fFileVar, GenSpace(LeftPos div Canvas.TextWidth(' '))+CurToken+GenSpace((CurWidth-(LeftPos+Canvas.TextWidth(CurToken))) div Canvas.TextWidth(' ') ) )
                else
                    Canvas.TextOut(fCurrentX+LeftPos, fCurrentY+StartingTop, CurToken);
                {Draw the cell lines if necessary.}
                if TableGrid then
                begin
                     X:=fCurrentX;
                     Y:=fCurrentY+StartingTop;
                     case LineSpacing of
                          lsHalfSpace: Y:=Y-(fTextMetrics.tmExternalLeading shr 1);
                          lsSingleSpace: Y:=Y-fTextMetrics.tmExternalLeading;
                          lsSingleAndAHalf: Y:=Y-Round(fTextMetrics.tmExternalLeading*1.5);
                          lsDoubleSpace: Y:=Y-(fTextMetrics.tmExternalLeading shl 1);
                     end;
                     {Draw the horizontal lines.}
                     Canvas.Pen.Width:=DefaultBorderWidth*(GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSY) div DefaultDPI);
                     Canvas.PolyLine([Point(X, Y), Point(X+CurWidth, Y)]);
                     Canvas.PolyLine([Point(X, Y+fLineSpace), Point(X+CurWidth, Y+fLineSpace)]);
                     {Draw the vertical lines.}
                     Canvas.Pen.Width:=DefaultBorderWidth*(GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSX) div DefaultDPI);
                     Canvas.PolyLine([Point(X, Y), Point(X, Y+fLineSpace)]);
                     Canvas.PolyLine([Point(X+CurWidth, Y), Point(X+CurWidth, Y+fLineSpace)]);
                end;
             except
                   on EPrinter do ;
             end;

             {Increase fCurrentX by the COLUMN width.}
             fCurrentX:=fCurrentX+CurWidth;
        end;
     finally
            FormatTokens.Free;
            LineTokens.Free;
     end;

     {If we're not printing the Header or Footer, go to a new line.}
     if ((fCurrentY >= 0) and
        (fCurrentY < PixelPrintHeight)) or
        fPrintingToFile then NewLine;
end;

procedure TLinePrinter.PrintLines(const LinesAsTable: Boolean);
var
   i: Integer; {This must allow negatives for Lines.Count-1}
begin
     for i:=0 to Lines.Count-1 do
     begin
          if LinesAsTable and (TableFormat<>'') then
             WriteTableLine(Lines[i])
          else
              WriteLine(Lines[i]);
     end;
     //Lines.Clear;
end;

procedure TLinePrinter.Refresh;
begin
     {This allows SetMarginXXX to make
     sure the margins are in bounds.

     It should be called after you display
     a printer setup dialog.  TLinePrinter
     can't detect a page size change, so this
     must be explicitly called to deal with
     any Margin boundary problems.}

     if not Printing then
     begin
          SetMarginTop(MarginTop);
          SetMarginBottom(MarginBottom);
          SetMarginLeft(MarginLeft);
          SetMarginRight(MarginRight);
     end;
end;

{=============================================================================}
{ Private and Protected stuff for TLinePrinter.                               }
{=============================================================================}

procedure TLinePrinter.DoNewPageProcessing;
var
   PixelBorderOffset: TPixels;
   OldTableFormat: String;
   OldTableGrid: Boolean;
begin
     UpdateProgressDlg('Currently Printing');

     {Keep TableFormat because we temporarily
     change it for the Header and Footer.
     Similar for TableGrid.}
     OldTableFormat:=TableFormat;
     OldTableGrid:=TableGrid;
     try
        TableGrid:=False;

        {Print the header.}
        if Header <> '' then
        begin
             {This value should be a negative offset.}
             fCurrentY:=((StartingTop-fLineSpace) shr 1)-StartingTop;
             TableFormat:=HeaderFormat;
             WriteTableLine(Header);
        end;

        {Print the footer.}
        if Footer <> '' then
        begin
             fCurrentY:=PixelPrintHeight+((StartingBottom-fLineSpace) shr 1);
             TableFormat:=FooterFormat;
             WriteTableLine(Footer);
        end;

     finally
            {Restore the original values.}
            TableFormat:=OldTableFormat;
            TableGrid:=OldTableGrid;
     end;

     {Reset the fields and fire new page and line events.}
     fCurrentX:=0;
     fCurrentY:=0;
     fLineNumber:=0;
     if Assigned(fOnNewPage) then fOnNewPage(Self);
     if Assigned(fOnNewLine) then fOnNewLine(Self);

     {Print the PageBorders.}
     try
        with Canvas do
        begin
             Pen.Width:=DefaultBorderWidth*(GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSY) div DefaultDPI);
             PixelBorderOffset:=MeasureUnitsToPixelsV(BorderOffset);
             if pbTop in PageBorders then
             begin
                  MoveTo(StartingLeft-PixelBorderOffset,StartingTop-PixelBorderOffset);
                  LineTo(StartingLeft+PixelPrintWidth+PixelBorderOffset, StartingTop-PixelBorderOffset);
             end;
             if pbBottom in PageBorders then
             begin
                  MoveTo(StartingLeft-PixelBorderOffset, StartingTop+PixelPrintHeight+PixelBorderOffset);
                  LineTo(StartingLeft+PixelPrintWidth+PixelBorderOffset, StartingTop+PixelPrintHeight+PixelBorderOffset);
             end;

             Pen.Width:=DefaultBorderWidth*(GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSX) div DefaultDPI);
             PixelBorderOffset:=MeasureUnitsToPixelsH(BorderOffset);
             if pbLeft in PageBorders then
             begin
                  MoveTo(StartingLeft-PixelBorderOffset, StartingTop-PixelBorderOffset);
                  LineTo(StartingLeft-PixelBorderOffset, StartingTop+PixelPrintHeight+PixelBorderOffset);
             end;
             if pbRight in PageBorders then
             begin
                  MoveTo(StartingLeft+PixelPrintWidth+PixelBorderOffset, StartingTop-PixelBorderOffset);
                  LineTo(StartingLeft+PixelPrintWidth+PixelBorderOffset, StartingTop+PixelPrintHeight+PixelBorderOffset);
             end;
        end;
     except
           on EPrinter do ;
     end;
end;

procedure TLinePrinter.SplitLineAndPrint(const Line: String);
var
   Buffer, CurLine: String;
   Len: Cardinal;
begin
     Buffer:=Line;
     repeat
           CurLine:=GetClippedLine(Buffer, PixelPrintWidth);
           Len:=Length(CurLine);
           {If the next character isn't whitespace, slide back to the nearest.
           Also, like most word processors do, I'm going to delete the
           first leading whitespace character left in the next-line buffer after
           the delete/newline (if one exists).}
           if Len<Length(Buffer) then
           begin
                if not (Buffer[Len+1] in [#0..#32]) then
                begin
                     CurLine:=StripBackToWhiteSpace(CurLine);
                     Len:=Length(CurLine);
                end
                else
                    Inc(Len);
           end;
           WriteLine(CurLine);
           Delete(Buffer, 1, Len);
     until Buffer='';
end;

function TLinePrinter.GetClippedLine(const Line: String; const Width: TPixels): String;
var
   PixelLen: TPixels;
   StartPos, EndPos, Mid, PreviousMid: Cardinal;
begin
     try
        PixelLen:=Canvas.TextWidth(Line);
        if PixelLen > Width then
        begin
             EndPos:=Length(Line);
             StartPos:=1;
             Mid:=0;
             repeat
                   PreviousMid:=Mid;
                   Mid:=(StartPos+EndPos) shr 1;
                   PixelLen:=Canvas.TextWidth(Copy(Line,1,Mid));

                   if PixelLen > Width then EndPos:=Mid
                   else if PixelLen < Width then StartPos:=Mid
                   else StartPos:=EndPos;
             until (Mid=PreviousMid) or (StartPos>=EndPos);
             Result:=Copy(Line, 1, Mid);
        end
        else
            Result:=Line;
     except
           on EPrinter do Result:=Line;
     end;
end;

function TLinePrinter.PixelPrintWidth: TPixels;
begin
     try
        Result:=MeasureUnitsToPixelsH(AvailablePageWidth)-StartingLeft-StartingRight;
     except
           on ERangeError do Result:=0;
     end;
end;

function TLinePrinter.PixelPrintHeight: TPixels;
begin
     try
        Result:=MeasureUnitsToPixelsV(AvailablePageHeight)-StartingTop-StartingBottom;
     except
           on ERangeError do Result:=0;
     end;
end;

function TLinePrinter.GetFont: TFont;
var
   Font: TFont;
begin
     Font:=TFont.Create;
     Font.Name:='Courier New';
     Font.Size:=10;
     Font.Style:=[];

     try
        Result:=Canvas.Font;
     except
           on EPrinter do
           begin
                Result:=Font;
           end;
     end;

     Font.Free;
end;

procedure TLinePrinter.SetFont(Value: TFont);
begin
     try
        Canvas.Font.Assign(Value);
        Canvas.Font.Size:=Value.Size;
        SetPixelsPerInch;
        if Printing and fPrintingToFile then
           GetTextMetrics(Canvas.Handle, fTextMetrics);
     except
           on EPrinter do
           begin
                if csDesigning in ComponentState then
                   MessageDlg('You must have at least one printer installed to set the Font property for TLinePrinter.',
                                   mtWarning, [mbOk], 0);
           end;
     end;
     {Force fLineSpace to be updated.}
     SetLineSpacing(LineSpacing);
end;

function TLinePrinter.GetTitle: String;
begin
     try
        Result:=fPrinter.Title;
     except
           on EPrinter do Result:='<Unknown>';
     end;
end;

procedure TLinePrinter.SetTitle(Value: String);
begin
     if not Printing then
        try
           fPrinter.Title:=Value
        except
              on EPrinter do
              begin
                   if csDesigning in ComponentState then
                      MessageDlg('You must have at least one printer installed to set the Title property for TLinePrinter.',
                                      mtWarning, [mbOk], 0);
              end;
        end
     else raise ELinePrinter.Create('Unable to change title while printing');
end;

function TLinePrinter.GetOrientation: TPrinterOrientation;
begin
     try
        Result:=fPrinter.Orientation;
     except
           on EPrinter do Result:=poPortrait;
     end;
end;

procedure TLinePrinter.SetOrientation(Value: TPrinterOrientation);
begin
     if not Printing then
        try
           fPrinter.Orientation:=Value
        except
              on EPrinter do
              begin
                   if csDesigning in ComponentState then
                      MessageDlg('You must have at least one printer installed to set the Orientation property for TLinePrinter.',
                                      mtWarning, [mbOk], 0);
              end;
        end
     else raise ELinePrinter.Create('Unable to change orientation while printing');
end;

function TLinePrinter.GetAborted: Boolean;
begin
     try
        Result:=fPrinter.Canceled;
     except
           on EPrinter do Result:=True;
     end;
end;

function TLinePrinter.GetAvailablePageHeight: TMeasurement;
begin
     try
        Result:=PixelsToMeasureUnitsV(fPrinter.PageHeight);
     except
        on EPrinter do
           if MeasureUnit = muInches then Result:=DefaultAvailablePageHeightIn
           else Result:=DefaultAvailablePageHeightMm;
     end;
end;

function TLinePrinter.GetPageNumber: Cardinal;
begin
     try
        if PrintToFile then
           Result:=fPageNumber
        else
            Result:=fPrinter.PageNumber;
     except
           on EPrinter do Result:=0;
     end;
end;

function TLinePrinter.GetAvailablePageWidth: TMeasurement;
begin
     try
        Result:=PixelsToMeasureUnitsH(fPrinter.PageWidth);
     except
        on EPrinter do
           if MeasureUnit = muInches then Result:=DefaultAvailablePageWidthIn
           else Result:=DefaultAvailablePageWidthMm;
     end;
end;

function TLinePrinter.GetPrinting: Boolean;
begin
     if PrintToFile then
        Result:=fPrintingToFile
     else
     begin
          if Assigned(fPrinter) then
             try
                Result:=(fPrinter.Status = mpPrinting);
             except
                   on EPrinter do Result:=False;
             end
          else Result:=False;
     end;
end;

procedure TLinePrinter.SetPixelsPerInch;
var
   FontSize: Integer;
begin
     {This routine gets us around the Delphi tiny font bug.}
     FontSize := Canvas.Font.Size;
     Canvas.Font.PixelsPerInch := GetDeviceCaps (Printer.Handle, LOGPIXELSY );
     Canvas.Font.Size := FontSize;
end;

procedure TLinePrinter.SetMarginTop(Value: TMeasurement);
begin
     if not Printing then
     begin
          if Value >= GutterTop then
          begin
               if Value <= (PhysicalPageHeight-GutterBottom) then fMarginTop:=Value
               else
               begin
                    fMarginTop:=PhysicalPageHeight-GutterBottom;
               end;
          end
          else
              fMarginTop:=GutterTop;
     end
     else raise ELinePrinter.Create('Unable to change top margin while printing');
end;

procedure TLinePrinter.SetMarginBottom(Value: TMeasurement);
begin
     if not Printing then
     begin
          if Value >= GutterBottom then
          begin
               if Value <= (PhysicalPageHeight-GutterTop) then fMarginBottom:=Value
               else
               begin
                    fMarginBottom:=PhysicalPageHeight-GutterTop;
               end;
          end
          else
              fMarginBottom:=GutterBottom;
     end
     else raise ELinePrinter.Create('Unable to change bottom margin while printing');
end;

procedure TLinePrinter.SetMarginLeft(Value: TMeasurement);
begin
     if not Printing then
     begin
          if Value >= GutterLeft then
          begin
               if Value <= (PhysicalPageWidth-GutterRight) then fMarginLeft:=Value
               else
               begin
                    fMarginLeft:=PhysicalPageWidth-GutterRight;
               end;
          end
          else
              fMarginLeft:=GutterLeft;
     end
     else raise ELinePrinter.Create('Unable to change left margin while printing');
end;

procedure TLinePrinter.SetMarginRight(Value: TMeasurement);
begin
     if not Printing then
     begin
          if Value >= GutterRight then
          begin
               if Value < (PhysicalPageWidth-GutterLeft) then fMarginRight:=Value
               else
               begin
                    fMarginRight:=PhysicalPageWidth-GutterLeft;
               end;
          end
          else
              fMarginRight:=GutterRight;
     end
     else raise ELinePrinter.Create('Unable to change right margin while printing');
end;

procedure TLinePrinter.SetMeasureUnit(Value: TMeasureUnit);
var
   OldMeas: TMeasureUnit;
begin
     OldMeas:=fMeasureUnit;
     fMeasureUnit:=Value;
     // Update the margins if the units have changed.
     if OldMeas <> MeasureUnit then
     begin
          if MeasureUnit = muInches then
          begin
               MarginTop:=MarginTop/25.4;
               MarginBottom:=MarginBottom/25.4;
               MarginLeft:=MarginLeft/25.4;
               MarginRight:=MarginRight/25.4;
               BorderOffset:=BorderOffset/25.4;
               DefaultColWidth:=DefaultColWidth/25.4;
          end
          else
          begin
               MarginTop:=MarginTop*25.4;
               MarginBottom:=MarginBottom*25.4;
               MarginLeft:=MarginLeft*25.4;
               MarginRight:=MarginRight*25.4;
               BorderOffset:=BorderOffset*25.4;
               DefaultColWidth:=DefaultColWidth*25.4;
          end;
     end;
     Refresh;
end;

procedure TLinePrinter.SetLineSpacing(Value: TLineSpacing);
var
   H: TPixels;
begin
     fLineSpacing:=Value;
     if Printing then
        H:=fTextMetrics.tmHeight+fTextMetrics.tmExternalLeading
     else
         try
            H:=Canvas.TextHeight('M');
         except
               on EPrinter do H:=10;
         end;

     case Value of
          lsHalfSpace: fLineSpace:=H shr 1;
          lsSingleSpace: fLineSpace:=H;
          lsSingleAndAHalf: fLineSpace:=H+(H shr 1);
          lsDoubleSpace: fLineSpace:=H+H;
     end;
end;

procedure TLinePrinter.SetTabSize(Value: Cardinal);
begin
     fTabSize:=Value;
end;

procedure TLinePrinter.SetWordWrap(Value: Boolean);
begin
     fWordWrap:=Value;
end;

function TLinePrinter.GetLines: TStrings;
begin
     if fLines = nil then
        fLines := TStringList.Create;
     GetLines := fLines;
end;

procedure TLinePrinter.SetLines(Value: TStrings);
begin
     if Value = nil then
     begin
          fLines.Free;
          fLines := nil;
          Exit;
     end;
     Lines.Assign(Value);
end;

procedure TLinePrinter.SetAlignment(Value: TAlignment);
begin
     fAlignment:=Value;
end;

procedure TLinePrinter.SetHeader(Value: String);
begin
     fHeader:=Value;
end;

procedure TLinePrinter.SetFooter(Value: String);
begin
     fFooter:=Value;
end;

procedure TLinePrinter.SetPageBorders(Value: TPageBorders);
begin
     fPageBorders:=Value;
end;

procedure TLinePrinter.SetShowProgress(Value: Boolean);
begin
     fShowProgress:=Value;
end;

procedure TLinePrinter.SetTableFormat(Value: String);
begin
     fTableFormat:=Value;
end;

procedure TLinePrinter.SetHeaderFormat(Value: String);
begin
     fHeaderFormat:=Value;
end;
procedure TLinePrinter.SetFooterFormat(Value: String);
begin
     fFooterFormat:=Value;
end;

procedure TLinePrinter.SetDefaultColWidth(Value: TMeasurement);
begin
     if Value >= 0 then
        fDefaultColWidth:=Value
     else
     begin
          fDefaultColWidth:=0;
          raise ELinePrinter.Create('DefaultColWidth must be greater than or equal to 0');
     end;
end;

procedure TLinePrinter.UpdateProgressDlg(const Status: String);
begin
     if ShowProgress and Printing then
     begin
          {Create it if is doesn't already exist.}
          if fPrntProgDlg = nil then
          begin
               fPrntProgDlg:=TPrntProgDlg.Create(Application);
          end;
          {Show it if it isn't visible.}
          if not fPrntProgDlg.Visible then fPrntProgDlg.Show;
          {Update it as necessary.}
          with fPrntProgDlg do
          begin
               if Status <> '' then lblStatus.Caption:=Status;
               lblTitle.Caption:=Title;
               lblPageNumber.Caption:=IntToStr(PageNumber);
               if Visible then Update;
          end;
     end
     else
     begin
          {If it exists, get rid of it.}
          if fPrntProgDlg <> nil then
          begin
               {If it is visible, close it.}
               if fPrntProgDlg.Visible then fPrntProgDlg.Close;
               fPrntProgDlg.Free;
               fPrntProgDlg := nil;
          end;
     end;
end;

function TLinePrinter.GetGutterTop: TMeasurement;
begin
     try
        Result:=PixelsToMeasureUnitsV(GetDeviceCaps(fPrinter.Canvas.Handle, PHYSICALOFFSETY));
     except
        on EPrinter do
           if MeasureUnit = muInches then Result:=DefaultGutterTopIn
           else Result:=DefaultGutterTopMm;
     end;
end;

function TLinePrinter.GetGutterBottom: TMeasurement;
begin
     Result:=PhysicalPageHeight-AvailablePageHeight-GutterTop;
end;

function TLinePrinter.GetGutterLeft: TMeasurement;
begin
     try
        Result:=PixelsToMeasureUnitsH(GetDeviceCaps(fPrinter.Canvas.Handle, PHYSICALOFFSETX));
     except
        on EPrinter do
           if MeasureUnit = muInches then Result:=DefaultGutterLeftIn
           else Result:=DefaultGutterLeftMm;
     end;
end;

function TLinePrinter.GetGutterRight: TMeasurement;
begin
     Result:=PhysicalPageWidth-AvailablePageWidth-GutterLeft;
end;

function TLinePrinter.StartingLeft: TPixels;
begin
     Result:=MeasureUnitsToPixelsH(MarginLeft-GutterLeft);
end;

function TLinePrinter.StartingRight: TPixels;
begin
     Result:=MeasureUnitsToPixelsH(MarginRight-GutterRight);
end;

function TLinePrinter.StartingTop: TPixels;
begin
     Result:=MeasureUnitsToPixelsV(MarginTop-GutterTop);
end;

function TLinePrinter.StartingBottom: TPixels;
begin
     Result:=MeasureUnitsToPixelsV(MarginBottom-GutterBottom);
end;

function TLinePrinter.MeasureUnitsToPixelsH(const M: TMeasurement): TPixels;
var
  Temp: TMeasurement;
begin
     Temp := M;
     try
        if MeasureUnit = muMillimeters then Temp := M / 25.4;
        Result:=Round((Temp*GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSX)));
     except
           on EPrinter do Result:=Round(Temp*DefaultDPI);
     end;
end;

function TLinePrinter.MeasureUnitsToPixelsV(const M: TMeasurement): TPixels;
var
  Temp: TMeasurement;
begin
     Temp := M;
     try
        if MeasureUnit = muMillimeters then Temp := M / 25.4;
        Result:=Round((Temp*GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSY)));
     except
           on EPrinter do Result:=Round(Temp*DefaultDPI);
     end;
end;

function TLinePrinter.PixelsToMeasureUnitsH(const P: TPixels): TMeasurement;
begin
     try
        Result:=(P / GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSX));
     except
           on EZeroDivide do Result:=P/DefaultDPI;
           on EPrinter do Result:=P/DefaultDPI;
     end;
     if MeasureUnit = muMillimeters then Result:=Result*25.4;
end;

function TLinePrinter.PixelsToMeasureUnitsV(const P: TPixels): TMeasurement;
begin
     try
        Result:=(P / GetDeviceCaps(fPrinter.Canvas.Handle, LOGPIXELSY));
     except
           on EZeroDivide do Result:=P/DefaultDPI;
           on EPrinter do Result:=P/DefaultDPI;
     end;
     if MeasureUnit = muMillimeters then Result:=Result*25.4;
end;

function TLinePrinter.ExpandLogicalFields(S: String): String;
begin
     S:=ReplaceSubString(LineField, IntToStr(LineNumber), S);
     S:=ReplaceSubString(PageField, IntToStr(PageNumber), S);
     S:=ReplaceSubString(DateField, FormatDateTime('ddddd',Date), S);
     S:=ReplaceSubString(TimeField, FormatDateTime('tt',Time), S);
     S:=ReplaceSubString(TitleField, Title, S);
     Result:=S;
end;

procedure TLinePrinter.SetBorderOffset(Value: TMeasurement);
begin
     fBorderOffset:=Value;
end;

function TLinePrinter.GetPhysicalPageHeight: TMeasurement;
begin
     try
        Result:=PixelsToMeasureUnitsV(GetDeviceCaps(fPrinter.Canvas.Handle, PHYSICALHEIGHT));
     except
        on EPrinter do
           if MeasureUnit = muInches then Result:=DefaultPhysicalPageHeightIn
           else Result:=DefaultPhysicalPageHeightMm;
     end;
end;

function TLinePrinter.GetPhysicalPageWidth: TMeasurement;
begin
     try
        Result:=PixelsToMeasureUnitsH(GetDeviceCaps(fPrinter.Canvas.Handle, PHYSICALWIDTH));
     except
        on EPrinter do
           if MeasureUnit = muInches then Result:=DefaultPhysicalPageWidthIn
           else Result:=DefaultPhysicalPageWidthMm;
     end;
end;

function TLinePrinter.GetPrintableWidth: TMeasurement;
begin
     Result:=PhysicalPageWidth-MarginLeft-MarginRight;
end;

function TLinePrinter.GetPrintableHeight: TMeasurement;
begin
     Result:=PhysicalPageHeight-MarginTop-MarginBottom;
end;

procedure TLinePrinter.SetPrintToFile(Value: Boolean);
begin
     if Value <> fPrintToFile then
     begin
          if not Printing then
          begin
               fPrintToFile:=Value;
          end
          else
              raise ELinePrinter.Create('Can''t change PrintToFile while printing.');
     end;
end;

procedure TLinePrinter.SetFileName(Value: String);
begin
     if Value <> fFileName then
     begin
          if not Printing then
          begin
               fFileName:=Trim(Value);
          end
          else
              raise ELinePrinter.Create('Can''t change FileName while printing.');
     end;
end;

procedure TLinePrinter.ExpandFriendlyFormat(UserFmt: String; AsHeader: Boolean);
var
   LayoutTokens: TStringList;
   NextCharIsSpecifier, TextAddOn: Boolean;
   i: Integer;
   FmtText, FmtLayout, Value: String;
begin
    LayoutTokens:=TStringList.Create;
    try
       NextCharIsSpecifier := False;
       FmtText:='';
       for i:=1 to Length(UserFmt) do
       begin
            //If it's an &, the next char is important.
            if (UserFmt[i] = '&') then
               NextCharIsSpecifier := True
            //The current char is a specifier of some sort.
            else if NextCharIsSpecifier then
            begin
                 Value := '';
                 TextAddOn := True;
                 case UserFmt[i] of
                     'f': Value := ExtractFileName(FileName);
                     'F': Value := FileName;
                     'd': Value := FormatDateTime('ddddd', Date);
                     't': Value := FormatDateTime('t', Time);
                     'D': Value := FormatDateTime('dddddd', Date);
                     'T': Value := FormatDateTime('tt', Time);
                     'p': Value := '{$PAGE}';
                     '&': Value := '&';
                     'l': begin Value := '<'; TextAddOn := False; end;
                     'c': begin Value := '^'; TextAddOn := False; end;
                     'r': begin Value := '>'; TextAddOn := False; end;
                 end;
                 //An empty filename will throw things off so we put in a space.
                 if Value = '' then Value:=' ';
                 
                 if TextAddOn then
                    FmtText:=FmtText+Value
                 else //We encountered an alignment specifier.
                 begin
                      if FmtText <> '' then
                      begin
                           if LayoutTokens.Count = 0 then LayoutTokens.Add('<');
                           FmtText:=FmtText+TokenSeparator;
                      end;
                      LayoutTokens.Add(Value);
                 end;
                 NextCharIsSpecifier := False;
            end
            //Otherwise we just add it on the Text string.
            else
                FmtText:=FmtText+UserFmt[i];
       end;

       //Now build the Layout string;
       FmtLayout:='';
       if LayoutTokens.Count = 0 then LayoutTokens.Add('<');
       Value := Trim(FloatToStrF((PrintableWidth / LayoutTokens.Count), ffFixed, 8, 2));
       for i:=0 to LayoutTokens.Count-1 do
       begin
            if FmtLayout <> '' then FmtLayout:=FmtLayout+TokenSeparator;
            FmtLayout:=FmtLayout+LayoutTokens[i]+Value;
       end;
    finally
           LayoutTokens.Free;
    end;

    //Now set the header or footer properties.
    if AsHeader then
    begin
         Header:=FmtText;
         HeaderFormat:=FmtLayout;
    end
    else
    begin
         Footer:=FmtText;
         FooterFormat:=FmtLayout;
    end;
end;

function TLinePrinter.GetFriendlyFooter: String;
begin
     //C++Builder 1.0 property bug workaround.
     Result:=fFriendlyFooter;
end;

procedure TLinePrinter.SetFriendlyFooter(Value: String);
begin
     fFriendlyFooter:=Value;
     ExpandFriendlyFormat(Value, False);
end;

function TLinePrinter.GetFriendlyHeader: String;
begin
     //C++Builder 1.0 property bug workaround.
     Result:=fFriendlyHeader;
end;

procedure TLinePrinter.SetFriendlyHeader(Value: String);
begin
     fFriendlyHeader:=Value;
     ExpandFriendlyFormat(Value, True);
end;

end.
